/*
 * Copyright (c) 2018, Ford Motor Company
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 *
 * Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following
 * disclaimer in the documentation and/or other materials provided with the
 * distribution.
 *
 * Neither the name of the Ford Motor Company nor the names of its contributors
 * may be used to endorse or promote products derived from this software
 * without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SRC_COMPONENTS_APPLICATION_MANAGER_INCLUDE_APPLICATION_MANAGER_HELP_PROMPT_MANAGER_H_
#define SRC_COMPONENTS_APPLICATION_MANAGER_INCLUDE_APPLICATION_MANAGER_HELP_PROMPT_MANAGER_H_

#include "smart_objects/smart_object.h"

namespace application_manager {
/**
 * @brief The HelpPromptManager is a mechanism for handling "help" requests
 * by the user when the application does not provide any or such information.
 */
class HelpPromptManager {
 public:
  /**
   * @brief Class destructor
   */
  virtual ~HelpPromptManager() {}

  /**
   * @brief Adds command to constructed values, and sends SetGlobalProperties if
   * required
   * @param cmd_id command unique ID
   * @param command smart object with commands to add
   * @param is_resumption flag for identifying if command was added during data
   * resumption process
   */
  virtual void OnVrCommandAdded(const uint32_t cmd_id,
                                const smart_objects::SmartObject& command,
                                const bool is_resumption) = 0;

  /**
   * @brief Removes command from constructed values, and send
   * SetGlobalProperties if required
   * @param cmd_id command unique ID
   * @param is_resumption flag for identifying if command was added during data
   * resumption process
   */
  virtual void OnVrCommandDeleted(const uint32_t cmd_id,
                                  const bool is_resumption) = 0;

  /**
   * @brief Stop constructing vrHelp and/or helpPrompt if they are present in
   * message
   * @param msg containing GlobalProperties
   * @param is_response determines is the request or response for the for
   * the SetGlobalPropertiesRequest
   */
  virtual void OnSetGlobalPropertiesReceived(
      const smart_objects::SmartObject& msg, const bool is_response) = 0;

  /**
   * @brief Triggered when ResetGlobalProperties request is received from an
   * application. Reset sending_type_ based on which global properties are reset
   * @param msg containing GlobalProperties
   */
  virtual void OnResetGlobalPropertiesReceived(
      const smart_objects::SmartObject& msg) = 0;

  /**
   * @brief Requests sending type behavior
   */
  enum class SendingType { kNoneSend, kSendHelpPrompt, kSendVRHelp, kSendBoth };

  /**
   * @brief Get current sending type
   * @return current sending type
   */
  virtual SendingType GetSendingType() const = 0;

  /**
   * @brief Construct the helpPrompt parameter
   */
  virtual void CreatePromptMsg(smart_objects::SmartObject& out_msg_params) = 0;

  /**
   * @brief Construct the vrHelp parameter
   */
  virtual void CreateVRMsg(smart_objects::SmartObject& out_msg_params) = 0;
};

}  //  namespace application_manager

#endif  // SRC_COMPONENTS_APPLICATION_MANAGER_INCLUDE_APPLICATION_MANAGER_HELP_PROMPT_MANAGER_H_
