/*

 Copyright (c) 2018, Ford Motor Company
 All rights reserved.

 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:

 Redistributions of source code must retain the above copyright notice, this
 list of conditions and the following disclaimer.

 Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following
 disclaimer in the documentation and/or other materials provided with the
 distribution.

 Neither the name of the Ford Motor Company nor the names of its contributors
 may be used to endorse or promote products derived from this software
 without specific prior written permission.

 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 */

#include "sdl_rpc_plugin/commands/mobile/reset_global_properties_request.h"

#include "application_manager/application_impl.h"
#include "application_manager/message_helper.h"
#include "interfaces/HMI_API.h"
#include "interfaces/MOBILE_API.h"

namespace sdl_rpc_plugin {
using namespace application_manager;

namespace commands {

SDL_CREATE_LOG_VARIABLE("Commands")

ResetGlobalPropertiesRequest::ResetGlobalPropertiesRequest(
    const application_manager::commands::MessageSharedPtr& message,
    ApplicationManager& application_manager,
    app_mngr::rpc_service::RPCService& rpc_service,
    app_mngr::HMICapabilities& hmi_capabilities,
    policy::PolicyHandlerInterface& policy_handler)
    : RequestFromMobileImpl(message,
                            application_manager,
                            rpc_service,
                            hmi_capabilities,
                            policy_handler)
    , ui_result_(hmi_apis::Common_Result::INVALID_ENUM)
    , tts_result_(hmi_apis::Common_Result::INVALID_ENUM)
    , rc_result_(hmi_apis::Common_Result::INVALID_ENUM) {}

ResetGlobalPropertiesRequest::~ResetGlobalPropertiesRequest() {}

void ResetGlobalPropertiesRequest::Run() {
  SDL_LOG_AUTO_TRACE();

  uint32_t app_id =
      (*message_)[strings::params][strings::connection_key].asUInt();
  ApplicationSharedPtr app = application_manager_.application(app_id);

  if (!app) {
    SDL_LOG_ERROR("No application associated with session key");
    SendResponse(false, mobile_apis::Result::APPLICATION_NOT_REGISTERED);
    return;
  }

  const auto& global_properties =
      (*message_)[strings::msg_params][strings::properties];

  auto reset_global_props_result =
      application_manager_.ResetGlobalProperties(global_properties, app_id);

  if (reset_global_props_result.HasUIPropertiesReset()) {
    StartAwaitForInterface(HmiInterfaces::HMI_INTERFACE_UI);
  }

  if (reset_global_props_result.HasTTSPropertiesReset()) {
    StartAwaitForInterface(HmiInterfaces::HMI_INTERFACE_TTS);
  }

  if (reset_global_props_result.HasRCPropertiesReset()) {
    StartAwaitForInterface(HmiInterfaces::HMI_INTERFACE_RC);
  }

  app->set_reset_global_properties_active(true);

  if (reset_global_props_result.HasUIPropertiesReset()) {
    // create ui request
    smart_objects::SmartObjectSPtr msg_params =
        MessageHelper::CreateUIResetGlobalPropertiesRequest(
            reset_global_props_result, app);
    if (msg_params.get()) {
      SendHMIRequest(
          hmi_apis::FunctionID::UI_SetGlobalProperties, msg_params.get(), true);
    }
  }

  if (reset_global_props_result.HasTTSPropertiesReset()) {
    smart_objects::SmartObjectSPtr msg_params =
        MessageHelper::CreateTTSResetGlobalPropertiesRequest(
            reset_global_props_result, app);

    SendHMIRequest(
        hmi_apis::FunctionID::TTS_SetGlobalProperties, msg_params.get(), true);
  }

  if (reset_global_props_result.HasRCPropertiesReset()) {
    smart_objects::SmartObjectSPtr msg_params =
        MessageHelper::CreateRCResetGlobalPropertiesRequest(
            reset_global_props_result, app);

    SendHMIRequest(
        hmi_apis::FunctionID::RC_SetGlobalProperties, msg_params.get(), true);
  }
  auto& help_prompt_manager = app->help_prompt_manager();
  help_prompt_manager.OnResetGlobalPropertiesReceived(
      (*message_)[strings::msg_params]);
}

void ResetGlobalPropertiesRequest::on_event(const event_engine::Event& event) {
  SDL_LOG_AUTO_TRACE();
  const smart_objects::SmartObject& message = event.smart_object();

  switch (event.id()) {
    case hmi_apis::FunctionID::UI_SetGlobalProperties: {
      SDL_LOG_INFO("Received UI_SetGlobalProperties event");
      EndAwaitForInterface(HmiInterfaces::HMI_INTERFACE_UI);
      ui_result_ = static_cast<hmi_apis::Common_Result::eType>(
          message[strings::params][hmi_response::code].asInt());
      GetInfo(message, ui_response_info_);
      break;
    }
    case hmi_apis::FunctionID::TTS_SetGlobalProperties: {
      SDL_LOG_INFO("Received TTS_SetGlobalProperties event");
      EndAwaitForInterface(HmiInterfaces::HMI_INTERFACE_TTS);
      tts_result_ = static_cast<hmi_apis::Common_Result::eType>(
          message[strings::params][hmi_response::code].asInt());
      GetInfo(message, tts_response_info_);
      break;
    }
    case hmi_apis::FunctionID::RC_SetGlobalProperties: {
      SDL_LOG_INFO("Received RC_SetGlobalProperties event");
      EndAwaitForInterface(HmiInterfaces::HMI_INTERFACE_RC);
      rc_result_ = static_cast<hmi_apis::Common_Result::eType>(
          message[strings::params][hmi_response::code].asInt());
      GetInfo(message, rc_response_info_);
      break;
    }
    default: {
      SDL_LOG_ERROR("Received unknown event " << event.id());
      return;
    }
  }

  if (IsPendingResponseExist()) {
    SDL_LOG_DEBUG("Waiting for remaining responses");
    return;
  }

  mobile_apis::Result::eType result_code = mobile_apis::Result::INVALID_ENUM;
  std::string response_info;
  const bool result = PrepareResponseParameters(result_code, response_info);

  SendResponse(result,
               static_cast<mobile_apis::Result::eType>(result_code),
               response_info.empty() ? NULL : response_info.c_str(),
               &(message[strings::msg_params]));
}

bool ResetGlobalPropertiesRequest::Init() {
  hash_update_mode_ = HashUpdateMode::kDoHashUpdate;
  return true;
}

bool ResetGlobalPropertiesRequest::PrepareResponseParameters(
    mobile_apis::Result::eType& out_result_code,
    std::string& out_response_info) {
  SDL_LOG_AUTO_TRACE();
  using namespace helpers;

  bool result = false;
  app_mngr::commands::ResponseInfo ui_properties_info(
      ui_result_, HmiInterfaces::HMI_INTERFACE_UI, application_manager_);
  app_mngr::commands::ResponseInfo tts_properties_info(
      tts_result_, HmiInterfaces::HMI_INTERFACE_TTS, application_manager_);
  app_mngr::commands::ResponseInfo rc_properties_info(
      rc_result_, HmiInterfaces::HMI_INTERFACE_RC, application_manager_);

  HmiInterfaces::InterfaceState tts_interface_state =
      application_manager_.hmi_interfaces().GetInterfaceState(
          HmiInterfaces::HMI_INTERFACE_TTS);

  if (hmi_apis::Common_Result::SUCCESS == ui_result_ &&
      hmi_apis::Common_Result::UNSUPPORTED_RESOURCE == tts_result_ &&
      HmiInterfaces::STATE_AVAILABLE == tts_interface_state) {
    result = true;
    out_result_code = mobile_apis::Result::WARNINGS;
    out_response_info = "Unsupported phoneme type sent in a prompt";
  } else {
    result = PrepareResultForMobileResponse(
        ui_properties_info, tts_properties_info, rc_properties_info);
    out_result_code = PrepareResultCodeForResponse(
        ui_properties_info, tts_properties_info, rc_properties_info);

    out_response_info = app_mngr::commands::MergeInfos(
        tts_response_info_, ui_response_info_, rc_response_info_);
  }

  return result;
}

}  // namespace commands

}  // namespace sdl_rpc_plugin
